# Copyright 2014, 2017, 2018 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

package MyRationals;
use 5.006;
use strict;
use Carp;
use base 'Exporter';
use Math::Matrix;

use Math::PlanePath::GcdRationals;
*gcd = \&Math::PlanePath::GcdRationals::_gcd;

our @EXPORT_OK = ('SB_left','SB_right',
                  'CW_left','CW_right',
                  'Bird_left','Bird_right',
                  'Drib_left','Drib_right',
                  'HCS_left','HCS_right',
                  'AYT_left','AYT_right',
                  'Kepler_left','Kepler_right',
                  'A1','A2','B1','B2','FibQ',
                  'Identity_2x2',
                  'mvec_to_pq','mvec_to_str',
                  'matrix_name',
                  'points_str',
                  'is_integer','is_pow2',
                  'pq_acceptable',
                  'xy_to_node_name',
                  'line_is_horizontal','line_is_diagonal','line_sort_ends',
                  'path_tree_to_hull_planar',
                  'gcd',
                  'planar_isboundary','planar_bigrat',
                  'F','Lucas',
                 );
our %EXPORT_TAGS = ( all => \@EXPORT_OK );

# uncomment this to run the ### lines
# use Smart::Comments;


use constant SB_left  => Math::Matrix->new ([1,0],
                                            [1,1]);
use constant SB_right => Math::Matrix->new ([1,1],
                                            [0,1]);
use constant CW_left  => SB_left();
use constant CW_right => SB_right();

use constant Bird_left  => Math::Matrix->new ([0,1],
                                              [1,1]);
use constant Bird_right => Math::Matrix->new ([1,1],
                                              [1,0]);
use constant Drib_left  => Bird_left();
use constant Drib_right => Bird_right();

use constant HCS_left  => Math::Matrix->new ([1,1],
                                             [0,1]);
use constant HCS_right => Math::Matrix->new ([0,1],
                                             [1,1]);
use constant AYT_left  => HCS_left();
use constant AYT_right => HCS_right();

use constant Kepler_left  => Math::Matrix->new ([1,0],
                                                [1,1]);
use constant Kepler_right => Math::Matrix->new ([0,1],
                                                [1,1]);

use constant Identity_2x2  => Math::Matrix->new ([1,0],
                                                 [0,1]);


use constant A1  => Math::Matrix->new ([1,0],
                                       [1,1]);
use constant A2  => Math::Matrix->new ([0,1],
                                       [1,1]);
use constant B1  => Math::Matrix->new ([1,1],
                                       [0,1]);
use constant B2  => Math::Matrix->new ([1,1],
                                       [1,0]);

use constant FibQ => Math::Matrix->new ([1,1],
                                        [1,0]);


#------------------------------------------------------------------------------

sub xy_to_node_name {
  my ($x,$y) = @_;
  if ($x==int($x) && $y==int($y)) {
    if ($x < 0) { $x = '-n'.-$x; }
    if ($y < 0) { $y = 'n'.-$y; }
    return "p$x-$y";
  } else {
    return "$x,$y";
  }
}


#------------------------------------------------------------------------------

use constant::defer F_ns => sub {
  require Math::NumSeq::Fibonacci;
  return Math::NumSeq::Fibonacci->new;
};
sub F {
  my ($n) = @_;
  return F_ns()->ith($n);
}

use constant::defer Lucas_ns => sub {
  require Math::NumSeq::LucasNumbers;
  return Math::NumSeq::LucasNumbers->new;
};
sub Lucas {
  my ($n) = @_;
  return Lucas_ns()->ith($n);
}

#------------------------------------------------------------------------------
# Math::Matrix things

# $m is a Math::Matrix which should be a 2x1 column vector [[$p],[$q]].
# Return the two values ($p,$q).
sub mvec_to_pq {
  my ($m) = @_;
  my ($rows,$cols) = $m->size;
  die unless $rows == 2 && $cols == 1;
  return ($m->[0][0], $m->[1][0]);
}

# $m is a Math::Matrix which should be a 2x1 column vector [[$p],[$q]].
# Return a string "$p,$q".
sub mvec_to_str {
  my ($m) = @_;
  my ($p,$q) = mvec_to_pq($m);
  return "$p,$q";
}


#------------------------------------------------------------------------------
# Math::Geometry::Planar things

# $points is an arrayref of [$x,$y] pairs.
# Return a string "x,y x,y ...".
sub points_str {
  my ($points) = @_;
  return  join(' ',map{join(',',@$_)}@$points)
}

#------------------------------------------------------------------------------
sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

BEGIN {
  my %matrix_name;

  my $mm_add = sub {
    my ($m, $name) = @_;
    $matrix_name{math_matrix_string($m)} ||= $name;
  };
  $mm_add->(A1,'A1');
  $mm_add->(A2,'A2');
  $mm_add->(B1,'B1');
  $mm_add->(B2,'B2');
  # $mm_add->(SB_left,'SB_left');
  # $mm_add->(SB_right,'SB_right');

  # products
  foreach my $pass (1, 2) {
    ### $pass
    ### %matrix_name
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      foreach my $y (@keys) {
        my ($a,$b,$c,$d) = split /,/, $y;
        my $y_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
        my $y_name = $matrix_name{$y};
        $mm_add->($x_2x2*$y_2x2, "$x_name.$y_name");
      }
    }
  }

  # inverses
  {
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      $mm_add->($x_2x2->invert, "($x_name)inv");
    }
  }

  # transposes
  {
    my @keys = keys %matrix_name;
    foreach my $x (@keys) {
      my ($a,$b,$c,$d) = split /,/, $x;
      my $x_2x2 = Math::Matrix->new([$a,$b],[$c,$d]);
      my $x_name = $matrix_name{$x};
      $mm_add->($x_2x2->transpose, "($x_name)transpose");
    }
  }

  sub math_matrix_string {
    my ($m) = @_;

    my ($width, $height) = $m->size;
    $width == 2 && $height == 2
      or $width == 3 && $height == 3
        or croak "math_matrix_string() not a 2x2 matrix";

    my @terms = map{@$_}@$m;
    return join(',',@terms);
  }

  sub matrix_name {
    my $key;
    if (@_ == 1) {
      my ($m) = @_;
      $key = math_matrix_string($m);
    } elsif (@_ == 4 || @_ == 9) {
      $key = join(',',@_);
    } else {
      croak "matrix_name() bad args";
    }
    return $matrix_name{$key} || '?';
  }

}

sub is_pow2 {
  my ($n) = @_;
  (($n >> 1) < $n) or die "is_pow2() infinite";
 
  while ($n > 1) {
    if ($n & 1) {
      return 0;
    }
    $n >>= 1;
  }
  return ($n == 1);
}

#------------------------------------------------------------------------------
# lines

# return true if the line x1,y1 to x2,y2 is horizontal, ie. $y1==$y2
sub line_is_horizontal {
  my ($x1,$y1, $x2,$y2) = @_;
  return $y1 == $y2;
}

# return true if the line x1,y1 to x2,y2 is diagonal, ie. slope +1
sub line_is_diagonal {
  my ($x1,$y1, $x2,$y2) = @_;
  return $x1-$y1 == $x2-$y2;
}

# return (x1,y1, x2,y2) with the ends swapped if necessary to make x1<x2, or
# if vertical x1=x2 then y1<y2
sub line_sort_ends {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2 || ($x1 == $x2 && $y1 > $y2)) {
    return ($x2,$y2, $x1,$y1);
  } else {
    return ($x1,$y1, $x2,$y2);
  }
}

#------------------------------------------------------------------------------

sub path_tree_to_hull_planar {
  my ($path, $depth, $type) = @_;
  $type ||= 'whole';
  my ($n_lo, $n_hi) = $path->tree_depth_to_n_range($depth);
  if ($type eq 'whole') {
    $n_lo = $path->n_start;
  }
  my $points = [];
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    push @$points, [$x,$y];
  }
  ### $points
  return points_to_hull_planar($points);
}

sub path_tree_depth_to_hull_area {
  my ($path, $depth) = @_;
  my $points = [ [1,1] ];
  foreach my $n ($path->n_start .. $path->tree_depth_to_n_end($depth)) {
    my ($x,$y) = $path->n_to_xy($n);
    push @$points, [$x,$y];
  }
  return points_to_hull_area($points);
}

sub points_to_hull_area {
  my ($points) = @_;
  if (@$points < 1) {
    return 0;
  }
  my $planar = points_to_hull_planar($points);
  return $planar->area;
}

sub points_to_hull_planar {
  my ($points) = @_;
  if (@$points < 3) {
    $points = [ $points->[0], $points->[0], @$points ];
  }
  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) { $planar = $planar->convexhull2; }
  return $planar;
}

#------------------------------------------------------------------------------
# Math::Geometry::Planar bits

# return true if $p is on the boundary of $planar
sub planar_isboundary {
  my ($planar, $p) = @_;
  ### planar_isboundary(): join(',',@$p).'  len='.scalar(@{$planar->points})
  my $points = $planar->points;
  if (@$points == 0) { return 0; }
  foreach my $i (0 .. $#$points) {
    my $p1 = $points->[$i-1];
    my $p2 = $points->[$i];
    ### consider: join(',',@$p1).' -- '.join(',',@$p2)
    if ((($p1->[0] <= $p->[0] && $p->[0] <= $p2->[0])
         || ($p2->[0] <= $p->[0] && $p->[0] <= $p1->[0]))
        && (($p1->[1] <= $p->[1] && $p->[1] <= $p2->[1])
            || ($p2->[1] <= $p->[1] && $p->[1] <= $p1->[1]))
        && Math::Geometry::Planar::Colinear([$p1, $p, $p2])) {
      ### found ...
      return 1;
    }
  }
  return 0;
}

# return a copy of $planar which has BigRat points
sub planar_bigrat {
  my ($planar) = @_;
  require Math::BigRat;
  my $points = $planar->points;
  my $new_points = [ map { [ Math::BigRat->new($_->[0]),
                             Math::BigRat->new($_->[1]) ]
                         } @$points ];
  my $new_planar = Math::Geometry::Planar->new;
  $new_planar->points($new_points);
  return $new_planar;
}


#------------------------------------------------------------------------------
1;
__END__
