# Copyright 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


package MyXYConvexHull;
use 5.010;
use strict;
use warnings;
use Carp 'croak';
use List::Util 'min','max','sum','sum0';
use MyDragon 'xy_rotate_minus90';

# uncomment this to run the ### lines
# use Smart::Comments;


sub new_from_points {
  my $class = shift;
  if (ref $class) { $class = ref $class; }

  my %x_to_yrange;
  my %y_to_xrange;
  foreach my $p (@_) {
    my ($x,$y) = @$p;
    _range_extend($x_to_yrange{$x}, $y);
    _range_extend($y_to_xrange{$y}, $x);
  }
  my $self = bless { x_to_yrange => \%x_to_yrange,
                     y_to_xrange => \%y_to_xrange,
                   }, $class;
  return _fill($self);
}
sub _range_extend {
  my ($aref, $n) = @_;
  unless ($aref) {
    $_[0] = [$n,$n];
    return 1;
  }
  if ($n < $aref->[0]) {
    $aref->[0] = $n;
    return 1;
  }
  if ($n > $aref->[1]) {
    $aref->[1] = $n;
    return 1;
  }
  return 0;
}
sub _fill {
  my ($self) = @_;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $y_to_xrange = $self->{'y_to_xrange'};
  my $more = 1;
  while ($more--) {
    while (my ($x,$yrange) = each %$x_to_yrange) {
      foreach my $y ($yrange->[0] .. $yrange->[1]) {
        $more |= _range_extend($y_to_xrange->{$y}, $x);
      }
    }
    while (my ($y,$xrange) = each %$y_to_xrange) {
      foreach my $x ($xrange->[0] .. $xrange->[1]) {
        $more |= _range_extend($x_to_yrange->{$x}, $y);
      }
    }
  }
  return $self->validate;
}
sub validate {
  my ($self) = @_;
  foreach my $href ($self->{'x_to_yrange'},
                    $self->{'y_to_xrange'}) {
    foreach my $range (values %$href) {
      ($range->[0] <= $range->[1])
        or croak "oops, range bad";
    }
  }
  return $self;
}

sub bbox {
  my ($self) = @_;
  my @xs = keys %{$self->{'x_to_yrange'}};
  my @ys = keys %{$self->{'y_to_xrange'}};
  if (@xs) {
    return (min(@xs),min(@ys), max(@xs),max(@ys));
  } else {
    return;
  }
}

sub x_to_yrange {
  my ($self, $y) = @_;
  my $aref = $self->{'x_to_yrange'} // return;
  return @$aref;
}
sub y_to_xrange {
  my ($self, $y) = @_;
  my $aref = $self->{'y_to_xrange'}->{$y} // return;
  return @$aref;
}

sub print {
  my ($self) = @_;
  my ($x1,$y1, $x2,$y2) = $self->bbox
    or return;
  my $y_to_xrange = $self->{'y_to_xrange'};

  my $ydigits = max(length($y1), length($y2));
  foreach my $y (reverse $y1 .. $y2) {
    printf " %*d  ", $ydigits, $y;
    my $xrange = $y_to_xrange->{$y};
    my ($xmin,$xmax) = @$xrange;
    print ' 'x($xmin-$x1);
    foreach my $x ($xmin .. $xmax) {
      if ($x==0 && $y==0) { print "O"; }
      else { print '*'; }
    }
    # print ' 'x($x2-$xmax);
    print "\n";
  }
  print "  points ",$self->num_points;
  print " (even ",$self->num_points_even;
  print ",odd ",$self->num_points_odd,")";
  my @num_sides = $self->num_sides4;
  print "  sides ",join(',',@num_sides)," = ",sum(@num_sides);
  print "\n";
}

sub num_points {
  my ($self) = @_;
  my $x_to_xrange = $self->{'x_to_yrange'};
  return sum0 (map {$_->[1]-$_->[0]+1} values %$x_to_xrange);
}

# return count of points in the hull which are on even $x+$y
sub num_points_even {
  my ($self) = @_;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $total = 0;
  while (my ($x,$yrange) = each %$x_to_yrange) {
    my ($ymin,$ymax) = @$yrange;
    ### column: "$x, $ymin to $ymax"
    if (($x+$ymin)&1) { $ymin++; }
    if (($x+$ymax)&1) { $ymax--; }
    $total += ($ymax - $ymin) / 2 + 1;
  }
  return $total;
}

# return count of points in the hull which are on odd $x+$y
sub num_points_odd {
  my ($self) = @_;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $total = 0;
  while (my ($x,$yrange) = each %$x_to_yrange) {
    my ($ymin,$ymax) = @$yrange;
    ### column: "$x, $ymin to $ymax"
    unless (($x+$ymin)&1) { $ymin++; }
    unless (($x+$ymax)&1) { $ymax--; }
    $total += ($ymax - $ymin) / 2 + 1;
  }
  return $total;
}

# return the area enclosed by the hull
sub area {
  my ($self) = @_;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my ($x1,$y1, $x2,$y2) = $self->bbox or return 0;
  my $area = 0;
  foreach my $y (reverse $y1 .. $y2-1) {
    my ($xmin1,$xmax1) = $self->y_to_xrange($y);
    my ($xmin2,$xmax2) = $self->y_to_xrange($y+1);
    $area += min($xmax1,$xmax2) - max($xmin1,$xmin2);
  }
  return $area;
}

sub num_sides {
  my ($self) = @_;
  return sum($self->num_sides4);
}
sub num_sides4 {
  my ($self) = @_;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $y_to_xrange = $self->{'y_to_xrange'};

  my @sides = (0,0,0,0);
  if (my ($x1,$y1, $x2,$y2) = $self->bbox) {
    @sides = (0,0,0,0);
    {
      ### vertical sides ...
      my $prev_y = $y1;
      my ($prev_xmin,$prev_xmax);
      foreach my $y ($y1 .. $y2) {
        my ($xmin,$xmax) = @{$y_to_xrange->{$y}};
        if (! defined $prev_xmin || $xmin != $prev_xmin) {
          $sides[2]++;
        }
        if (! defined $prev_xmax || $xmax != $prev_xmax) {
          $sides[0]++;
        }
        ($prev_xmin,$prev_xmax) = ($xmin,$xmax);
      }
    }
    {
      ### horizontal sides ...
      my $prev_x = $x1;
      my ($prev_ymin,$prev_ymax);
      foreach my $x ($x1 .. $x2) {
        my ($ymin,$ymax) = @{$x_to_yrange->{$x}};
        ### this y range: "$ymin, $ymax"
        if (! defined $prev_ymin || $ymin != $prev_ymin) {
          $sides[3]++;
          ### new S: $sides[3]
        }
        if (! defined $prev_ymax || $ymax != $prev_ymax) {
          $sides[1]++;
          ### new N: $sides[1]
        }
        ($prev_ymin,$prev_ymax) = ($ymin,$ymax);
      }
    }
  }
  return @sides;
}


sub rotate_plus90 {
  my ($self) = @_;
  return $self->rotate_minus90->rotate_minus90->rotate_minus90;
}
sub rotate_180 {
  my ($self) = @_;
  return $self->rotate_minus90->rotate_minus90;
}
sub rotate_minus90 {
  my ($self) = @_;
  my $class = ref $self;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $y_to_xrange = $self->{'y_to_xrange'};
  my %new_x_to_yrange;
  my %new_y_to_xrange;
  while (my ($x,$yrange) = each %$x_to_yrange) {
    (undef, my $new_y) = xy_rotate_minus90($x,0);
    (my $new_x1, undef) = xy_rotate_minus90(0, $yrange->[0]);
    (my $new_x2, undef) = xy_rotate_minus90(0, $yrange->[1]);
    $new_y_to_xrange{$new_y} = [ min($new_x1,$new_x2), max($new_x1,$new_x2) ];
  }
  while (my ($y,$xrange) = each %$y_to_xrange) {
    (my $new_x, undef) = xy_rotate_minus90(0,$y);
    (undef, my $new_y1) = xy_rotate_minus90($xrange->[0], 0);
    (undef, my $new_y2) = xy_rotate_minus90($xrange->[1], 0);
    $new_x_to_yrange{$new_x} = [ min($new_y1,$new_y2), max($new_y1,$new_y2) ];
  }
  $self = bless { x_to_yrange => \%new_x_to_yrange,
                  y_to_xrange => \%new_y_to_xrange,
                }, ref $self;
  return $self->validate;
}

sub translate {
  my ($self, $dx,$dy) = @_;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $y_to_xrange = $self->{'y_to_xrange'};
  my %new_x_to_yrange;
  my %new_y_to_xrange;
  while (my ($x,$yrange) = each %$x_to_yrange) {
    $new_x_to_yrange{$x+$dx} = [ $yrange->[0]+$dy, $yrange->[1]+$dy ];
  }
  while (my ($y,$xrange) = each %$y_to_xrange) {
    $new_y_to_xrange{$y+$dy} = [ $xrange->[0]+$dx, $xrange->[1]+$dx ];
  }
  $self = bless { x_to_yrange => \%new_x_to_yrange,
                  y_to_xrange => \%new_y_to_xrange,
                }, ref $self;
  return $self->validate;
}

sub _range_union {
  my ($r1, $r2) = @_;
  $r1 //= $r2;
  return [ min($r1->[0], $r2->[0]), max($r1->[1], $r2->[1]) ];
}

sub union {
  my $self = shift;
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $y_to_xrange = $self->{'y_to_xrange'};
  $x_to_yrange = { %$x_to_yrange };
  $y_to_xrange = { %$y_to_xrange };
  foreach my $other (@_) {
    my $other_x_to_yrange = $other->{'x_to_yrange'};
    my $other_y_to_xrange = $other->{'y_to_xrange'};
    while (my ($other_x,$other_yrange) = each %$other_x_to_yrange) {
      $x_to_yrange->{$other_x} = _range_union($x_to_yrange->{$other_x},
                                              $other_yrange);
    }
    while (my ($other_y,$other_xrange) = each %$other_y_to_xrange) {
      $y_to_xrange->{$other_y} = _range_union($y_to_xrange->{$other_y},
                                              $other_xrange);
    }
  }
  $self = bless { x_to_yrange => $x_to_yrange,
                  y_to_xrange => $y_to_xrange,
                }, ref $self;
  return _fill($self);
}

sub vertices {
  my ($self) = @_;
  my ($x1,$y1, $x2,$y2) = $self->bbox or return [];
  my $x_to_yrange = $self->{'x_to_yrange'};
  my $y_to_xrange = $self->{'y_to_xrange'};

  my @points;
  my $prev_y;
  my $prev_x;
  my $to = sub {
    my ($x,$y) = @_;
    ### to: "$x,$y"
    if (@points >= 1) {
      if ($x==$prev_x && $y==$prev_y) {
        ### is same ...
        return;
      }
    }
    if (@points >= 2) {
      if ($x==$prev_x && $x==$points[-2]->[0]
          && ($y <=> $prev_y) == ($prev_y <=> $points[-2]->[1])) {
        ### extend vertical: "$points[-1]->[0], $y"
        $points[-1]->[1] = $y;
        $prev_x = $x;
        $prev_y = $y;
        return;
      }
      if ($y==$prev_y && $y==$points[-2]->[1]
          && ($x <=> $prev_x) == ($prev_x <=> $points[-2]->[0])) {
        ### extend horizontal ...
        $points[-1]->[0] = $x;
        $prev_x = $x;
        $prev_y = $y;
        return;
      }
    }
    push @points, [$x,$y];
    $prev_x = $x;
    $prev_y = $y;
  };

  # right
  $to->($y_to_xrange->{$y1}->[1], $y1);
  foreach my $y ($y1 .. $y2) {
    my $x = $y_to_xrange->{$y}->[1];
    ### right: "y=$y  x=$x  prev_x=$prev_x"
    if ($x > $prev_x) {
      $to->($prev_x, $y);
    } elsif ($x < $prev_x) {
      $to->($x, $prev_y);
    }
    $to->($x, $y);
  }

  # left
  $to->($y_to_xrange->{$y2}->[0], $y2);
  foreach my $y (reverse $y1 .. $y2) {
    my $x = $y_to_xrange->{$y}->[0];
    if ($x < $prev_x) {
      $to->($prev_x, $y);
    } elsif ($x > $prev_x) {
      $to->($x, $prev_y);
    }
    $to->($x, $y);
  }

  return \@points;
}

#------------------------------------------------------------------------------
1;
__END__
