#!/usr/bin/perl -w

# Copyright 2014, 2017, 2018, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Print a tikz picture showing the points on the outer boundary of the
# convex hull around the SB tree.

use 5.010;
use strict;
use List::Util 'max','min';
use Math::PlanePath::RationalsTree;
use Math::Geometry::Planar;

use FindBin;
use lib $FindBin::Bin;
use MyRationals 'gcd';
use MyPrintwrap ':all';

# types:
#     outer
#     row
#     whole
my $type = 'whole';
if (@ARGV) { $type = $ARGV[0]; }

my $depth = 5;

my $path = Math::PlanePath::RationalsTree->new(tree_type => 'SB');

print "  % generated by tools/$FindBin::Script $type\n";

if ($type eq 'outer') {
  $Printwrap = 2;
  printwrap("% row lines\n");
  printwrap("\\draw[ultra thin] ");
  $Printwrap += 2;
  my ($n_row, $n_end) = $path->tree_depth_to_n_range($depth);
  foreach my $n ($n_row .. $n_end) {
    my ($x,$y) = $path->n_to_xy($n);
    my $join = ($n == $n_row ? '' : '--');
    my $end =  ($n == $n_end ? ';' : '');
    printwrap("$join($x,$y)$end");
  }
  printwrap_newline();
}

my $points = [];  # arrayref of point arrayrefs [$x,$y]
my $n_row = ($type eq 'row' ? $path->tree_depth_to_n($depth)
             : $path->n_start);
my $n_end = $path->tree_depth_to_n_end($depth);
foreach my $n ($n_row .. $n_end) {
  my ($x,$y) = $path->n_to_xy($n);
  push @$points, [$x,$y];
}
my $planar = Math::Geometry::Planar->new;
$planar->points($points);
if (@$points > 3) { $planar = $planar->convexhull2; }
$points = $planar->points;

{
  $Printwrap = 2;
  printwrap("% hull boundary\n");
  printwrap("\\draw[thin] ");
  $Printwrap += 2;
  my $join = '';

  my $cycle = ($type eq 'outer' ? '' : "--cycle;");
  my $start = ($type eq 'outer' ? 2
               : $type eq 'row' ? 0
               :                  0);
  foreach my $i ($start .. $#$points) {
    my $p = $points->[$i];
    my ($x,$y) = @$p;
    my $end = (! $cycle && $i == $#$points ? ';' : '');
    printwrap("$join($x,$y)$end");
    $join = '--';
  }
  printwrap($cycle);
  printwrap_newline();
}

# dots for integer points on boundary
{
  $Printwrap = 2;
  printwrap("% hull boundary integer points\n");
  my $count = 0;
  foreach my $i (0 .. $#$points) {
    my $p1 = $points->[$i-1];
    my $p2 = $points->[$i];
    my ($x1,$y1) = @$p1;
    my ($x2,$y2) = @$p2;
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;
    my $g = gcd($dx,$dy);
    $dx /= $g;
    $dy /= $g;
    printwrap("  % $x1,$y1 to $x2,$y2  dx=$dx dy=$dy\n");
    my $x = $x1;
    my $y = $y1;
    for (;
         ! ($x == $x2 && $y == $y2);
         $x += $dx, $y += $dy) {
      my $style = (($x == $x1 && $y == $y1)
                   || ($x == $x2 && $y == $y2)
                   ? '[big dot]'
                   : '[small dot]');
      if ($type eq 'outer') {
        next if $x + $y < $depth+2;
      }
      printwrap("  \\fill$style ($x,$y) circle;");
      $count++;
    }
    printwrap_newline();
  }
  printwrap("  % count $count\n");
}
print "  % end generated\n";

exit 0;
