#!/usr/bin/perl -w

# Copyright 2013, 2014, 2015, 2016, 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl test.pl
#
# Run some self-tests on ../rationals.tex and on MyRationals.pm.
#

use 5.010;
use strict;
use File::Slurp;
use IPC::Run;
use List::Util 'max','min';
use Math::Matrix;
use Math::PlanePath::RationalsTree;
use Math::NumSeq::Fibonacci;
use Math::Geometry::Planar;

use FindBin;
use lib::abs $FindBin::Bin;
use MyRationals ':all';

use Memoize;
BEGIN { memoize('path_tree_to_hull_planar'); }

# uncomment this to run the ### lines
# use Smart::Comments;


my $source_dir = File::Spec->catdir ($FindBin::Bin, File::Spec->updir);
$source_dir = File::Spec->rel2abs($source_dir);
print "source dir $source_dir\n";

chdir $source_dir or die $!;

my $rationals_tex_filename = 'rationals.tex';
# $rationals_tex_filename = 'devel/more.tex';
my $rationals_pdf_filename = 'rationals.pdf';
my $rationals_tex_str = File::Slurp::slurp($rationals_tex_filename);
print count_lines($rationals_tex_str)," lines\n";

my $rationals_tex_str_sans_comments = $rationals_tex_str;
$rationals_tex_str_sans_comments =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
print count_lines($rationals_tex_str_sans_comments)," lines\n";

my $sb_path = Math::PlanePath::RationalsTree->new;


#------------------------------------------------------------------------------
# ../devel/rationals-defines.gp
# ../devel/rationals-defines.header.gp

{
  # same copyright years as rationals.tex where defines are extracted from
  $rationals_tex_str =~ /^% (Copyright .*)/m or die;
  my $want_copyright = $1;
  foreach my $filename ('devel/rationals-defines.gp',
                        'devel/rationals-defines.header.gp') {
    my $str = File::Slurp::slurp($filename);
    $str =~ /^\\\\ \Q$want_copyright\E$/m
      or die "$filename copyright different from rationals.tex";
  }
}

#------------------------------------------------------------------------------
# convex hull centroid

# FIXME: Math::Geometry::Planar starts some sums from undefs which
# Math::BigRat doesn't like.
if (0) {
  # HV(d,i) = [ F(i+1); F(i+2) + (d-i)*F(i+1) ]
  #
  my $count = 0;
  foreach my $d (2 .. 10) {
    my $planar = HG_by_planar($d);
    my $formula = HG_by_formula($d);
    $planar == $formula
      or die "hull centroid d=$d planar $planar formula $formula";
    $count++;
  }
  print "$count convex hull centroid\n";

  sub HG_by_planar {
    my ($d, $type) = @_;
    if ($d == 0) { return 1; }
    my $planar = path_tree_to_hull_planar($sb_path, $d, 'whole');
    $planar = planar_bigrat($planar);
    my $centroid = $planar->centroid;
    return $centroid->[0];
    my $rect = $planar->minrectangle;
    return $rect->area;
  }
  sub HG_by_formula {
    my ($d) = @_;
    if ($d == 0) { return 1; }
    require Math::BigRat;
    $d = Math::BigRat->new($d);
    return (F(3*$d+7) - 4*$d*$d - 22*$d -13 ) / 24 / H_by_formula($d);
  }
}


#------------------------------------------------------------------------------
# minrectangle

# hull formulas, whole tree
foreach my $d (0 .. 10) {
  my $AMR_formula = AMR_by_formula($d);
  my $AMR_planar = AMR_by_planar($d);
  $AMR_formula == $AMR_planar
    or die "minrect formula $AMR_formula planar $AMR_planar";

  my $BMR_formula = BMR_by_formula($d);
  my $BMR_planar = BMR_by_planar($d);
  $BMR_formula == $BMR_planar
    or die "minrect BMR";

  print "hull d=$d  AMR $AMR_formula $AMR_planar  BMR $BMR_formula $BMR_planar\n";
}

sub AMR_by_formula {
  my ($d) = @_;
  return (F($d+2) - 1) ** 2;
}
sub AMR_by_planar {
  my ($d, $type) = @_;
  if ($d == 0) { return 0; }
  my $planar = path_tree_to_hull_planar($sb_path, $d, 'whole');
  $planar = planar_bigrat($planar);
  my $rect = $planar->minrectangle;
  return $rect->area;
}

sub BMR_by_formula {
  my ($d) = @_;
  return 4*(F($d+2) - 1);
}
sub BMR_by_planar {
  my ($d, $type) = @_;
  if ($d == 0) { return 0; }
  my $planar = path_tree_to_hull_planar($sb_path, $d, 'whole');
  $planar = planar_bigrat($planar);
  my $rect = $planar->minrectangle;
  return $rect->perimeter;
}


#------------------------------------------------------------------------------
# tree area centroid

# FIXME: Math::Geometry::Planar starts some sums from undefs which
# Math::BigRat doesn't like.
if (0) {
  foreach my $d (0 .. 6) {
    my $by_formula = G_by_formula($d);
    my $by_planar = G_by_planar($d);
    print "d=$d G $by_planar $by_formula\n";
    $by_planar == $by_formula or die;
  }
}

sub G_by_planar {
  my ($d) = @_;
  if ($d == 0) { return 1; }
  my $path = Math::PlanePath::RationalsTree->new (tree_type => 'SB');
  my @points = ([1,1]);
  my ($n_row, $n_end) = $path->tree_depth_to_n_range($d);
  foreach my $n ($n_row .. $n_end) {
    my ($x,$y) = $path->n_to_xy($n);
    push @points, [$x,$y];
  }
  my $planar = Math::Geometry::Planar->new;
  $planar->points(\@points);
  $planar = planar_bigrat($planar);
  return $planar->centroid->[0];
}

sub G_by_formula {
  my ($d) = @_;
  if ($d == 0) { return 1; }
  return GN_by_formula($d) / R_by_formula($d);
}

sub GN_by_formula {
  my ($d) = @_;
  # (19.3^d - 6d^2 - 24d - 19) / 12
  require Math::BigRat;
  $d = Math::BigRat->new($d);
  return (19*3**$d - 6*$d*$d - 24*$d - 19) / 12;
}

sub R_by_formula {
  my ($d) = @_;
  # \frac{5}{2} (2^d-1) - 2d
  require Math::BigRat;
  $d = Math::BigRat->new($d);
  return (2**$d - 1)*5/2 - 2*$d;
}


#------------------------------------------------------------------------------
# convex hull

sub H_by_formula {
  my ($depth) = @_;
  return (F(2*$depth+3)/2 - $depth - 1);
}

sub H_by_formula_2 {
  my ($depth) = @_;
  return (F($depth+2) ** 2
          - F($depth+0) ** 2 / 2
          - F($depth+1) * F($depth)
          - $depth
          - 1);

  return ((F($depth+2) - 1) ** 2
          - F($depth+0) ** 2 / 2
          - 2 * (F($depth+1)*F($depth)/2
                 - F($depth+2) + $depth/2 + 1));
}

sub H_by_planar {
  my ($depth, $type) = @_;
  my $planar = path_tree_to_hull_planar($sb_path, $depth, $type);
  return $planar->area;
}

# hull formulas, whole tree
foreach my $depth (0 .. 10) {
  my $formula = H_by_formula($depth);
  my $formula2 = H_by_formula_2($depth);
  my $planar = H_by_planar($depth,'whole');
  print "hull $depth   $planar $formula\n";
  $formula == $planar or die "formula $formula planar $planar";
  $formula2 == $planar or die "formula2 $formula2 planar $planar";
}

print "# H_by_planar(row)\n";
foreach my $depth (0 .. 10) {
  # print "\\tfrac{",2*H_by_planar($depth,'row'),"}{2},\n";
  print 2*H_by_planar($depth,'row'),"/2, ";
}
print "\n";

# extra area of 0,0 alone
foreach my $d (0 .. 10) {
  my $planar = Math::Geometry::Planar->new;
  $planar->points([ [0,0], [$d+1,1], [1,1], [1,$d+1] ]);
  my $area = $planar->area;
  my $want = $d;
  $area == $want or die "d=$d got area=$area want $want";
}

# extra area of 0,1 and 1,0
foreach my $d (0 .. 10) {
  my $planar = Math::Geometry::Planar->new;
  $planar->points([ [1,0], [$d+1,1], [1,1], [1,$d+1], [0,1] ]);
  my $area = $planar->area;
  my $want = $d + 1/2;
  $area == $want or die "d=$d got area=$area want $want";
}

# extra area of 0,1 and 1,0 and 0,0
foreach my $d (0 .. 10) {
  my $planar = Math::Geometry::Planar->new;
  $planar->points([ [1,0], [$d+1,1], [1,1], [1,$d+1], [0,1], [0,0] ]);
  my $area = $planar->area;
  my $want = $d + 1;
  $area == $want or die "d=$d got area=$area want $want";
}


#------------------------------------------------------------------------------
# check MY-BEGIN-MATRICES

{
  while ($rationals_tex_str
         =~ /MY-BEGIN-MATRICES ([^\n]+)(.*?)MY-END-MATRICES/sg) {
    my $names = $1;
    my $content = $2;
    my $linenum = pos_linenum($rationals_tex_str);
    my @names = split /,/, $names;

    my $i = 0;
    while ($content =~ /begin\{(p?(small)?matrix)\}(.*)?end\{(p?(small)?matrix)\}/sg) {
      my $pmatrix = $3;
      ### $pmatrix
      ### $i

      my @got_terms = split m{[^-0-9/]+}, $pmatrix;
      if ($got_terms[0] eq '') { shift @got_terms } # leading empty

      # 1/2 -> 0.5
      foreach my $t (@got_terms) {
        if ($t =~ m{(.*)/2$}) {
          $t = $1 / 2;  # mutate array
        }
      }

      ### @got_terms
      my $got_terms = join(',',@got_terms);

      my $want_name = $names[$i%3];
      @got_terms == 4
        or @got_terms == 9
          or die "rationals.tex:$linenum: MY-BEGIN-MATRICES $want_name terms: ",scalar(@got_terms),"\n";

      my $got_name = matrix_name(@got_terms);
      $got_name eq $want_name
        or die "rationals.tex:$linenum: MY-BEGIN-MATRICES $got_terms want $want_name got $got_name\n";

      print "MY-BEGIN-MATRICES $want_name ok, line $linenum\n";
      $i++;
    }
    if ($i == 0) {
      die "rationals.tex:$linenum: MY-BEGIN-MATRICES pmatrix not matched\n";
    }
  }
}

#------------------------------------------------------------------------------
# Benson samples cf Kepler FractionsTree

{
  require Math::PlanePath::FractionsTree;
  my $path = Math::PlanePath::FractionsTree->new;
  foreach my $elem (['1011', 1,2, 2,3, 2,5, 5,7, 7,12],
                    ['1101', 1,2, 2,3, 3,5, 3,8, 8,11],
                    ['1010000', 1,2, 2,3, 2,5, 5,7, 5,12, 5,17, 5,22,  5,27],
                    ['0000101', 1,2, 1,3, 1,4, 1,5,  1,6,  6,7, 6,13, 13,19],
                   ) {
    my ($bits, @want_xy_list) = @$elem;
    scalar(@want_xy_list) == 2*(length($bits)+1)
      or die "not enough xy list";
    foreach my $len (0 .. length($bits)) {
      my $want_x = shift @want_xy_list;
      my $want_y = shift @want_xy_list;
      my $n_bits = substr($bits, 0, $len);
      my $n = oct "0b1$n_bits";
      my ($got_x,$got_y) = $path->n_to_xy($n)
        or die "nothing at n=$n";
      my $want_str = "$want_x,$want_y";
      my $got_str = "$got_x,$got_y";
      $want_str eq $got_str
        or die sprintf "len=$len n=%b want $want_str got $got_str", $n;
    }
  }
}


#------------------------------------------------------------------------------
# check MY-PARI-CHECK expressions

{
  my $pari_preload = '';

  $pari_preload .= "sbL=".matrix_to_pari_str(SB_left()).";\n";
  $pari_preload .= "sbR=".matrix_to_pari_str(SB_right()).";\n";
  #
  $pari_preload .= "cwL=".matrix_to_pari_str(CW_left()).";\n";
  $pari_preload .= "cwR=".matrix_to_pari_str(CW_right()).";\n";

  $pari_preload .= "birdL=".matrix_to_pari_str(Bird_left()).";\n";
  $pari_preload .= "birdR=".matrix_to_pari_str(Bird_right()).";\n";
  #
  $pari_preload .= "dribL=".matrix_to_pari_str(Drib_left()).";\n";
  $pari_preload .= "dribR=".matrix_to_pari_str(Drib_right()).";\n";

  $pari_preload .= "hcsL=".matrix_to_pari_str(HCS_left()).";\n";
  $pari_preload .= "hcsR=".matrix_to_pari_str(HCS_right()).";\n";
  #
  $pari_preload .= "aytL=".matrix_to_pari_str(AYT_left()).";\n";
  $pari_preload .= "aytR=".matrix_to_pari_str(AYT_right()).";\n";

  $pari_preload .= "keplerL=".matrix_to_pari_str(Kepler_left()).";\n";
  $pari_preload .= "keplerR=".matrix_to_pari_str(Kepler_right()).";\n";

  $pari_preload .= "a1=".matrix_to_pari_str(A1()).";\n";
  $pari_preload .= "b1=".matrix_to_pari_str(B1()).";\n";
  $pari_preload .= "a2=".matrix_to_pari_str(A2()).";\n";
  $pari_preload .= "b2=".matrix_to_pari_str(B2()).";\n";
  $pari_preload .= "fibQ=".matrix_to_pari_str(FibQ()).";\n";

  print "Pari preload:\n$pari_preload\n";

  while ($rationals_tex_str =~ /MY-PARI-(CHECK|DEFINE):?[ \t](.*)/mg) {
    my $type = $1;
    my $pari_input = $2;
    if ($type eq 'DEFINE') {
      $pari_preload .= "$pari_input\n";
      next;
    }
    my $linenum = pos_linenum($rationals_tex_str);
    print "$pari_input\n";

    $pari_input = $pari_preload . $pari_input;
    my $pari_output;
    IPC::Run::run(['gp','--quiet'],
                  '<', \$pari_input,
                  '>', \$pari_output)
        or die;
    unless ($pari_output =~ /^1\Z/m) {
      print $pari_output;
      print "rationals.tex:$linenum: MY-PARI-CHECK expected 1\n";
      print "\n";
      print "full input:\n$pari_input\n";
      die;
    }
  }

  sub matrix_to_pari_str {
    my ($m) = @_;
    return '['. join(';', map{join(',',@$_)} @$m) . ']';
  }
}


#------------------------------------------------------------------------------
# descents
#        tree  HCS
#       /    \
# tree+1    1/(tree + 1)

sub path_descent_str {
  my ($path, $n, $depth, $m) = @_;
  my $ret = '';
  my $depth_limit = $path->tree_n_to_depth($n) + $depth;
  my @pending = ($n);
  while (@pending) {
    my $n = shift @pending;
    my $depth = $path->tree_n_to_depth($n);
    next if $depth > $depth_limit;

    my ($x,$y) = $path->n_to_xy($n) or die;
    ($x,$y) = mvec_to_pq($m * Math::Matrix->new([$x],[$y]));
    $ret .= "$x,$y ";
    push @pending, $path->tree_n_children($n);
  }
  return $ret;
}

foreach my $elem (['SB',   SB_left(),  SB_right()],
                  ['HCS',  HCS_left(), HCS_right()],
                  ['Bird', Bird_left(),Bird_right()]) {
  my ($tree_type, $L, $R) = @$elem;

  my $path = Math::PlanePath::RationalsTree->new(tree_type=>$tree_type);
  my $depth = 10;
  my $n_start = $path->n_start;
  my ($n_left,$n_right) = $path->tree_n_children($n_start);

  {
    my $want_left = path_descent_str($path, $n_left, $depth, Identity_2x2());
    my $got_left = path_descent_str($path, $n_start, $depth, $L);
    ### $want_left
    ### $got_left
    $want_left eq $got_left or die;
    print "$tree_type left len ",length($want_left),"\n";
  }
  {
    my $want_right = path_descent_str($path, $n_right, $depth, Identity_2x2());
    my $got_right = path_descent_str($path, $n_start, $depth, $R);
    ### $want_right
    ### $got_right
    $want_right eq $got_right or die;
    print "$tree_type right len ",length($want_right),"\n";
  }
}


#------------------------------------------------------------------------------
# MyRationals.pm matrices vs Math::PlanePath::RationalsTree

foreach my $elem (['CW',   CW_left(),  CW_right()],
                  ['AYT',  AYT_left(), AYT_right()],
                  ['Drib', Drib_left(),Drib_right()]) {
  my ($tree_type, $L, $R) = @$elem;
  my $path = Math::PlanePath::RationalsTree->new(tree_type=>$tree_type);
  foreach my $n ($path->n_start .. 1000) {
    my ($x,$y) = $path->n_to_xy($n) or die;
    my ($n_left,$n_right) = $path->tree_n_children($n);
    my ($x_left,$y_left) = $path->n_to_xy($n_left);
    my ($x_right,$y_right) = $path->n_to_xy($n_right);

    {
      my $v = Math::Matrix->new([$x],[$y]);
      $v = $L * $v;
      my ($x_mvec,$y_mvec) = mvec_to_pq($v);
      $x_mvec == $x_left or die;
      $y_mvec == $y_left or die;
    }
    {
      my $v = Math::Matrix->new([$x],[$y]);
      $v = $R * $v;
      my ($x_mvec,$y_mvec) = mvec_to_pq($v);
      $x_mvec == $x_right or die;
      $y_mvec == $y_right or die;
    }
  }
}

#------------------------------------------------------------------------------
# check MY-BEGIN-PQ-POINTS

{
  while ($rationals_tex_str
         =~ /MY-BEGIN-PQ-POINTS(.*?)MY-END-PQ-POINTS/sg) {
    my $content = $1;
    my $linenum = pos_linenum($rationals_tex_str);

    my (@p,@q);
    my %seen_pq;
    while ($content =~ m{\((\d+), *(\d+)\)}g) {
      my $p = $1;
      my $q = $2;
      pq_acceptable($p,$q)
        or die "rationals.tex:$linenum: MY-BEGIN-PQ-POINTS bad point $p,$q\n";
      push @p, $p;
      push @q, $q;
      if ($seen_pq{"$p,$q"}) {
        die "rationals.tex:$linenum: MY-BEGIN-PQ-POINTS duplicate $p,$q\n";
      }
      $seen_pq{"$p,$q"} = 1;
    }

    my $max_p = max(@p);
    foreach my $p (2 .. $max_p) {
      foreach my $q (1 .. $p-1) {
        next unless pq_acceptable($p,$q);
        $seen_pq{"$p,$q"}
          or die "rationals.tex:$linenum: MY-BEGIN-PQ-POINTS missing point $p,$q\n";
      }
    }
    print "MY-BEGIN-PQ-POINTS ok ",scalar(@p)," points, line $linenum\n";
  }
}

#------------------------------------------------------------------------------
# rationals.pdf Title and Author

{

  my $pdfinfo_out;
  IPC::Run::run(['pdfinfo', '-isodates', $rationals_pdf_filename],
                '>', \$pdfinfo_out)
      or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;
  {
    $pdfinfo_out =~ /^Title:\s*(.*)/m
      or die;
    my $title = $1;
    # print "title: $title\n";
    $title =~ /Trees/
      or die;
  }
  {
    $pdfinfo_out =~ /^Author:\s*(.*)/m
      or die;
    my $author = $1;
    # print "author: $author\n";
    $author =~ /Kevin/
      or die;
  }
}

#------------------------------------------------------------------------------
sub pos_linenum {
  my ($str) = @_;
  $str = substr($str, 0, pos($_[0]));
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

#------------------------------------------------------------------------------
exit 0;
